import enum
import random
import time
from typing import List

import dill
import numpy as np
import shap
from deap import base
from deap import creator
from deap import gp
from deap import tools
from deap.gp import PrimitiveTree, Primitive, Terminal
from numpy.testing import assert_almost_equal
from sklearn import model_selection
from sklearn.base import RegressorMixin, ClassifierMixin
from sklearn.datasets import make_regression
from sklearn.feature_selection import VarianceThreshold
from sklearn.inspection import permutation_importance
from sklearn.linear_model import RidgeCV, LogisticRegression
from sklearn.linear_model._base import LinearModel
from sklearn.metrics import make_scorer, accuracy_score, balanced_accuracy_score, precision_score, recall_score, \
    f1_score, r2_score
from sklearn.model_selection import StratifiedKFold, cross_validate, train_test_split, KFold
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.tree import BaseDecisionTree

from evolutionary_forest.model.PLTree import SoftPLTreeRegressor, SoftPLTreeRegressorEM
from evolutionary_forest.multigene_gp import result_post_process, MultiplePrimitiveSet, quick_fill
from evolutionary_forest.sklearn_utils import cross_val_predict
from evolutionary_forest.utils import reset_random

reset_random(0)


def calculate_score(args):
    (X, Y, score_func, cv, other_parameters) = calculate_score.data
    # nn_prediction: prediction results of the neural network
    nn_prediction = other_parameters['nn_prediction']
    # dynamic_target: random generate cross-validation scheme
    dynamic_target = other_parameters['dynamic_target']
    original_features = other_parameters['original_features']
    # test_data_size: number of test data samples, which do not have the label
    test_data_size = other_parameters['test_data_size']
    pset = other_parameters['pset']
    # sklearn_format: model is in sklearn format
    sklearn_format = other_parameters['sklearn_format']
    cross_validation = other_parameters['cross_validation']
    feature_importance_method = other_parameters['feature_importance_method']
    filter_elimination = other_parameters['filter_elimination']

    pipe: Pipeline
    pipe, func = args
    if not isinstance(func, list):
        func = dill.loads(func)

    # only used for PS-Tree
    best_label = None
    altered_labels = None
    # GP evaluation
    start_time = time.time()

    if sklearn_format:
        Yp = quick_result_calculation(func, pset, X, original_features, sklearn_format)
        pipe = pipe_combine(Yp, pipe)
        Yp = X
        hash_result = None
    else:
        if hasattr(pipe, 'register'):
            Yp, hash_result = quick_result_calculation(func, pset, X, original_features, need_hash=True,
                                                       register_array=pipe.register)
        else:
            Yp, hash_result = quick_result_calculation(func, pset, X, original_features, need_hash=True)
        if nn_prediction is not None:
            Yp = np.concatenate([Yp, nn_prediction], axis=1)
        if test_data_size > 0:
            Yp = Yp[:-test_data_size]
        assert isinstance(Yp, np.ndarray)
        assert not np.any(np.isnan(Yp))
        assert not np.any(np.isinf(Yp))

        # only use for PS-Tree
        if hasattr(pipe, 'partition_scheme'):
            partition_scheme = pipe.partition_scheme
            assert not np.any(np.isnan(partition_scheme))
            assert not np.any(np.isinf(partition_scheme))
            Yp = np.concatenate([Yp, np.reshape(partition_scheme, (-1, 1))], axis=1)
    if hasattr(pipe, 'active_gene'):
        Yp = Yp[:, pipe.active_gene]

    gp_evaluation_time = time.time() - start_time

    # ML evaluation
    start_time = time.time()
    if 'CV' in score_func:
        # custom cross validation scheme
        if '-Random' in score_func:
            cv = StratifiedKFold(shuffle=True)
        else:
            cv = None

        if score_func == 'CV-Accuracy-Recall':
            custom_scorer = {'accuracy': make_scorer(accuracy_score),
                             'balanced_accuracy': make_scorer(balanced_accuracy_score),
                             'precision': make_scorer(precision_score, average='macro'),
                             'recall': make_scorer(recall_score, average='macro'),
                             'f1': make_scorer(f1_score, average='macro')}
            result = cross_validate(pipe, Yp, Y, cv=cv, return_estimator=True, scoring=custom_scorer)
        elif score_func == 'CV-F1-Score':
            result = cross_validate(pipe, Yp, Y, cv=cv, return_estimator=True,
                                    scoring=make_scorer(f1_score, average='macro'))
        elif score_func == 'CV-SingleFold':
            result = single_fold_validation(pipe, Yp, Y, index=random.randint(0, 4))
        else:
            result = cross_validate(pipe, Yp, Y, cv=cv, return_estimator=True)

        if score_func == 'CV-Accuracy-Recall':
            y_pred = np.array([np.mean(result['test_accuracy']),
                               np.mean(result['test_balanced_accuracy']),
                               np.mean(result['test_precision']),
                               np.mean(result['test_recall']),
                               np.mean(result['test_f1'])])
        else:
            y_pred = result['test_score']
        estimators = result['estimator']
    elif not cross_validation:
        pipe.fit(Yp, Y)
        y_pred = pipe.predict(Yp)
        estimators = [pipe]
    else:
        if sklearn_format:
            base_model = pipe['model']['Ridge']
        else:
            base_model = pipe['Ridge']
        regression_task = isinstance(base_model, RegressorMixin)
        if isinstance(base_model, RidgeCV):
            # Ridge CV
            pipe.fit(Yp, Y)
            y_pred, estimators = base_model.cv_values_[:, np.argmax(base_model.cv_values_.sum(axis=0))], [pipe]
        elif cv == 1:
            # single fold training (not recommend)
            indices = np.arange(len(Y))
            x_train, x_test, y_train, y_test, idx_train, idx_test = train_test_split(
                Yp, Y, indices, test_size=0.2)
            estimators = [pipe]
            pipe.fit(x_train, y_train)
            y_pred = np.ones_like(Y)
            y_pred[idx_test] = pipe.predict(x_test)
        else:
            # cross-validation
            if dynamic_target:
                cv = get_cv_splitter(base_model, cv, random.randint(0, int(1e9)))
            else:
                cv = get_cv_splitter(base_model, cv)

            if filter_elimination is not None:
                strategies = set(filter_elimination.split(','))
                mask = np.ones(X.shape[1], dtype=bool)
                if 'Variance' in strategies:
                    # eliminate features based on variance
                    threshold = VarianceThreshold(threshold=0.01)
                    threshold.fit(Yp)
                    indices = threshold.get_support(indices=True)
                    mask[indices] = False
                if 'Correlation' in strategies:
                    # eliminate features based on correlation
                    col_corr = []
                    corr_matrix = np.corrcoef(Yp)
                    for i in range(corr_matrix.shape[1]):
                        for j in range(i):
                            if abs(corr_matrix[i, j]) > threshold:
                                # highly correlated features
                                col_corr.append(i)
                                break
                    mask[col_corr] = False
                Yp = Yp[:, mask]

            if regression_task:
                y_pred, estimators = cross_val_predict(pipe, Yp, Y, cv=cv)
            else:
                y_pred, estimators = cross_val_predict(pipe, Yp, Y, cv=cv, method='predict_proba')

            if feature_importance_method == 'SHAP' and len(estimators) == cv.n_splits:
                for id, estimator in enumerate(estimators):
                    split_fold = list(cv.split(Yp, Y))
                    train_id, test_id = split_fold[id][0], split_fold[id][1]
                    if isinstance(estimator['Ridge'], (LinearModel, LogisticRegression)):
                        explainer = shap.LinearExplainer(estimator['Ridge'], Yp[train_id])
                    elif isinstance(estimator['Ridge'], BaseDecisionTree):
                        explainer = shap.TreeExplainer(estimator['Ridge'], Yp[train_id])
                    else:
                        raise Exception
                    if isinstance(estimator['Ridge'], BaseDecisionTree):
                        feature_importance = explainer.shap_values(Yp[test_id])[0]
                    else:
                        feature_importance = explainer.shap_values(Yp[test_id])
                    estimator['Ridge'].shap_values = np.abs(feature_importance).mean(axis=0)

            if feature_importance_method == 'PermutationImportance' and len(estimators) == cv.n_splits:
                # Don't need to calculate the mean value here
                for id, estimator in enumerate(estimators):
                    split_fold = list(cv.split(Yp, Y))
                    train_id, test_id = split_fold[id][0], split_fold[id][1]
                    r = permutation_importance(estimator['Ridge'], Yp[test_id], Y[test_id], n_jobs=1, n_repeats=1)
                    estimator['Ridge'].pi_values = np.abs(r.importances_mean)

            if np.any(np.isnan(y_pred)):
                np.save('error_data_x.npy', Yp)
                np.save('error_data_y.npy', Y)
                raise Exception

            if isinstance(base_model, SoftPLTreeRegressor) and not isinstance(base_model, SoftPLTreeRegressorEM):
                # determine the best partition scheme
                cv_label = other_parameters['cv_label']
                if cv_label:
                    best_label = np.zeros(len(Y))
                else:
                    partition_num = len(np.unique(pipe.partition_scheme))
                    best_label = np.zeros((len(Y), partition_num))
                for index, estimator in zip(cv.split(Y), estimators):
                    estimator: SoftPLTreeRegressor
                    if cv_label:
                        # determine the label of PS-Tree through cross-validation
                        train_index, test_index = index
                        X_test, y_test = Yp[test_index], Y[test_index]
                        best_label[test_index] = estimator.score(X_test, np.reshape(y_test, (-1, 1)))
                    else:
                        train_index, test_index = index
                        score = estimator.score(Yp[train_index], np.reshape(Y[train_index], (-1, 1)))
                        best_label[train_index, score] += 1
                if not cv_label:
                    best_label = np.argmax(best_label, axis=1)
                altered_labels = np.sum(pipe.partition_scheme != best_label)
    ml_evaluation_time = time.time() - start_time
    if altered_labels is not None:
        print('altered_labels', altered_labels)
    return y_pred, estimators, {
        'gp_evaluation_time': gp_evaluation_time,
        'ml_evaluation_time': ml_evaluation_time,
        'best_label': best_label,
        'altered_labels': altered_labels,
        'hash_result': hash_result,
    }


def pipe_combine(Yp, pipe):
    Yp = list(filter(lambda x: not isinstance(x, (int, float, np.ndarray, enum.Enum)), Yp))
    if len(Yp) == 0:
        pipe = Pipeline([
            ("feature", "passthrough"),
            ('model', pipe)
        ])
    else:
        pipe = Pipeline([
            ("feature", FeatureUnion(
                [(f'preprocessing_{id}', p) for id, p in enumerate(Yp)]
            )),
            ('model', pipe)
        ])
    return pipe


def single_fold_validation(model, x_data, y_data, index):
    # only validate single fold, this method is faster than cross-validation
    cv = model_selection.KFold(n_splits=5, shuffle=False, random_state=0)
    current_index = 0
    scores = []
    for train_index, test_index in cv.split(x_data):
        if current_index == index:
            # print("TRAIN:", train_index, "TEST:", test_index)
            X_train, X_test = x_data[train_index], x_data[test_index]
            y_train, y_test = y_data[train_index], y_data[test_index]
            model.fit(X_train, y_train)
            if isinstance(model['Ridge'], ClassifierMixin):
                scores.append(accuracy_score(y_test, model.predict(X_test)))
            elif isinstance(model['Ridge'], RegressorMixin):
                scores.append(r2_score(y_test, model.predict(X_test)))
            else:
                raise Exception
        else:
            scores.append(-1)
        current_index += 1
    assert np.any(scores != 0)
    return {
        'test_score': scores,
        'estimator': [model for _ in range(5)],
    }


def get_cv_splitter(base_model, cv, random_state=0):
    if isinstance(base_model, ClassifierMixin):
        cv = StratifiedKFold(n_splits=cv, shuffle=True, random_state=random_state)
    else:
        cv = KFold(n_splits=cv, shuffle=True, random_state=random_state)
    return cv


def quick_result_calculation(func: List[PrimitiveTree], pset, data, original_features=False,
                             sklearn_format=False, need_hash=False, register_array=None):
    if sklearn_format:
        data = enum.Enum('Enum', {f"ARG{i}": i for i in range(data.shape[1])})
    # quick evaluate the result of features
    result = []
    hash_result = []
    if hasattr(pset, 'number_of_register'):
        register = np.ones((data.shape[0], pset.number_of_register))
        for id, gene in enumerate(func):
            input_data = np.concatenate([data, register], axis=1)
            quick_result = quick_evaluate(gene, pset, input_data)
            quick_result = quick_fill([quick_result], data)[0]
            if isinstance(quick_result, np.ndarray):
                hash_result.append(hash(quick_result.tostring()))
            else:
                hash_result.append(hash(str(quick_result)))
            register[:, register_array[id]] = quick_result
        result = register.T
    elif isinstance(pset, MultiplePrimitiveSet):
        for id, gene in enumerate(func):
            quick_result = quick_evaluate(gene, pset.pset_list[id], data)
            quick_result = quick_fill([quick_result], data)[0]
            if isinstance(quick_result, np.ndarray):
                hash_result.append(hash(quick_result.tostring()))
            else:
                hash_result.append(hash(str(quick_result)))
            result.append(quick_result)
            data = np.concatenate([data, np.reshape(quick_result, (-1, 1))], axis=1)
    else:
        for gene in func:
            feature: np.ndarray = quick_evaluate(gene, pset, data)
            if isinstance(feature, np.ndarray):
                hash_result.append(hash(feature.tostring()))
            else:
                hash_result.append(hash(str(feature)))
            result.append(feature)
    if not sklearn_format:
        result = result_post_process(result, data, original_features)
        # result = np.reshape(result[:, -1], (-1, 1))
    if not need_hash:
        return result
    else:
        return result, hash_result


def quick_evaluate(expr: PrimitiveTree, pset, data, prefix='ARG'):
    result = None
    stack = []
    for node in expr:
        stack.append((node, []))
        while len(stack[-1][1]) == stack[-1][0].arity:
            prim, args = stack.pop()
            if isinstance(prim, Primitive):
                result = pset.context[prim.name](*args)
            elif isinstance(prim, Terminal):
                if prefix in prim.name:
                    if isinstance(data, np.ndarray):
                        result = data[:, int(prim.name.replace(prefix, ''))]
                    elif isinstance(data, (dict, enum.EnumMeta)):
                        result = data[prim.name]
                    else:
                        raise ValueError("Unsupported data type!")
                else:
                    result = prim.value
            else:
                raise Exception
            if len(stack) == 0:
                break  # If stack is empty, all nodes should have been seen
            stack[-1][1].append(result)
    return result


def minimal_task():
    # A minimal task for SR
    x, y = make_regression(n_samples=1000)
    pset = gp.PrimitiveSet("MAIN", x.shape[1])
    pset.addPrimitive(np.add, 2, name="vadd")
    pset.addPrimitive(np.subtract, 2, name="vsub")
    pset.addPrimitive(np.multiply, 2, name="vmul")
    pset.addPrimitive(np.negative, 1, name="vneg")
    pset.addPrimitive(np.cos, 1, name="vcos")
    pset.addPrimitive(np.sin, 1, name="vsin")
    pset.addEphemeralConstant("rand101", lambda: random.randint(-1, 1))
    creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
    creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMin)
    toolbox = base.Toolbox()
    toolbox.register("expr", gp.genHalfAndHalf, pset=pset, min_=2, max_=8)
    toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.expr)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("compile", gp.compile, pset=pset)
    pop = toolbox.population(n=10000)
    st = time.time()
    avg_a = np.zeros(x.shape[0])
    for ind in pop:
        avg_a += quick_evaluate(ind, pset, x)
    print('time', time.time() - st)
    st = time.time()
    avg_b = np.zeros(x.shape[0])
    for ind in pop:
        func = gp.compile(ind, pset)
        avg_b += func(*x.T)
    print('time', time.time() - st)
    assert_almost_equal(avg_a, avg_b)


def minimal_feature_importance():
    from sklearn.datasets import load_diabetes
    from sklearn.linear_model import Ridge

    diabetes = load_diabetes()
    X_train, X_val, y_train, y_val = train_test_split(
        diabetes.data, diabetes.target, random_state=0)
    model = Ridge(alpha=1e-2).fit(X_train, y_train)
    model.score(X_val, y_val)
    r = permutation_importance(model, X_val, y_val,
                               n_repeats=30,
                               random_state=0)
    print(r.importances_mean)
    for i in r.importances_mean.argsort()[::-1]:
        if r.importances_mean[i] - 2 * r.importances_std[i] > 0:
            print(f"{diabetes.feature_names[i]:<8}"
                  f"{r.importances_mean[i]:.3f}"
                  f" +/- {r.importances_std[i]:.3f}")


if __name__ == '__main__':
    # minimal_task()
    minimal_feature_importance()



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.base import TransformerMixin
from sklearn.datasets import make_classification
from skorch import NeuralNet
from skorch.callbacks import EarlyStopping
from skorch.utils import to_numpy
from torch import optim
from torch.nn import MSELoss


class VAE(nn.Module):
    def __init__(self, input_unit, hidden_layer=64, embedding_size=2):
        super(VAE, self).__init__()
        self.input_unit = input_unit
        self.hidden_layer = hidden_layer
        self.embedding_size = embedding_size

        self.fc1 = nn.Linear(input_unit, hidden_layer)
        self.fc21 = nn.Linear(hidden_layer, embedding_size)
        self.fc22 = nn.Linear(hidden_layer, embedding_size)
        self.fc3 = nn.Linear(embedding_size, hidden_layer)
        self.fc4 = nn.Linear(hidden_layer, input_unit)

    def encode(self, x):
        h1 = F.relu(self.fc1(x))
        return self.fc21(h1), self.fc22(h1)

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return eps.mul(std).add_(mu)
        else:
            return mu

    def decode(self, z):
        h3 = F.relu(self.fc3(z))
        return self.fc4(h3).view(-1, self.input_unit)

    def forward(self, x):
        mu, logvar = self.encode(x.view(-1, self.input_unit))
        z = self.reparameterize(mu, logvar)
        return self.decode(z), mu, logvar


class NeuralNetTransformer(NeuralNet, TransformerMixin):
    def get_loss(self, y_pred, y_true, X, **kwargs):
        recons, mu, log_var = y_pred
        self.beta = 0.1
        recons_loss = F.mse_loss(recons, X)
        kld_loss = torch.mean(-0.5 * torch.sum(1 + log_var - mu ** 2 - log_var.exp(), dim=1), dim=0)
        loss = recons_loss + self.beta * kld_loss
        return loss

    def transform(self, X):
        X = X.astype(np.float32)
        out = []
        for outs in self.forward_iter(X, training=False):
            outs = outs[1] if isinstance(outs, tuple) else outs
            out.append(to_numpy(outs))
        transforms = np.concatenate(out, 0)
        return transforms


if __name__ == '__main__':
    X, y = make_classification(1000, 20, n_informative=10, random_state=0)
    X, y = X.astype(np.float32), y.astype(np.int64)

    vae = NeuralNetTransformer(VAE,
                               criterion=MSELoss(),
                               optimizer=optim.Adam,
                               module__input_unit=X.shape[1],
                               max_epochs=1000,
                               callbacks=[EarlyStopping(patience=20)],
                               verbose=True)
    vae.fit(X, y)
    print(vae.transform(X))



from abc import abstractmethod
from typing import Dict, List, Union, Callable

import numpy as np
from mlxtend.classifier import LogisticRegression
from scipy.special import softmax
from sklearn.base import ClassifierMixin, RegressorMixin, BaseEstimator
from sklearn.cluster import KMeans, DBSCAN
from sklearn.datasets import load_wine, load_diabetes
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import BaggingRegressor
from sklearn.linear_model import RidgeCV, LinearRegression, Ridge, LogisticRegression
from sklearn.linear_model._base import LinearModel
from sklearn.metrics import r2_score, roc_auc_score
from sklearn.mixture import GaussianMixture
from sklearn.mixture._base import BaseMixture
from sklearn.model_selection import train_test_split, KFold
from sklearn.multioutput import MultiOutputRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder, StandardScaler, OneHotEncoder
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier, BaseDecisionTree


class PLTree(BaseDecisionTree):
    @abstractmethod
    def model_controller(self):
        pass

    def post_fit(self, X, y, partition_scheme=None):
        # construct linear models in each partition
        self.model_map: Dict = {}
        model_coefs = []
        if partition_scheme is None:
            labels = super().apply(X)
            for l in np.unique(labels):
                idx = labels == l
                model = self.model_controller()
                if len(np.unique(y[idx])) == 1:
                    model = DummyClassifier()
                    model.coef_ = np.zeros(X.shape[1])
                    model = model.fit(X[idx], y[idx])
                else:
                    model = model.fit(X[idx], y[idx])
                self.model_map[l] = model
                # model.predict_proba(X)
                if len(model.coef_.shape) == 2:
                    model_coefs.append(np.max(model.coef_, axis=0))
                else:
                    model_coefs.append(model.coef_)
        else:
            for l, p in enumerate(partition_scheme.T):
                model = self.model_controller()
                index = p > 0
                model = model.fit(X[index], y[index], sample_weight=p[index])
                self.model_map[l] = model
                model_coefs.append(model.coef_)

        lr_coef = np.abs(np.array(model_coefs)).max(axis=0)
        lr_coef /= np.sum(lr_coef)
        self.feature_importance = np.max([lr_coef, self.feature_importances_[:len(lr_coef)]], axis=0)
        self.feature_importance /= np.sum(self.feature_importance)
        assert len(self.feature_importance) == X.shape[1]

    def post_predict(self, X, labels, method=None, classes=None):
        classes_map = {}
        if method == 'predict_proba':
            # for classification task, we need to predict probabilities
            predictions = np.zeros((X.shape[0], self.n_classes_))
            classes_map = {
                v: k
                for k, v in enumerate(self.classes_)
            }
        elif method == 'soft_prediction':
            if isinstance(self.partition_tree, DecisionTreeClassifier):
                predictions = np.zeros((X.shape[0], self.partition_tree.n_classes_))
            elif isinstance(self.partition_tree, LogisticRegression):
                predictions = np.zeros((X.shape[0], len(self.partition_tree.classes_)))
            elif isinstance(self.partition_tree, GaussianMixture):
                n_classes = len(np.unique(self.partition_tree.classes_))
                predictions = np.zeros((X.shape[0], n_classes))
            else:
                raise Exception
        else:
            predictions = np.zeros(X.shape[0])

        if method == 'soft_prediction':
            # get all prediction results
            for l in classes:
                model = self.model_map[l]
                predictions[:, l] = model.predict(X)
        else:
            for l in np.sort(np.unique(labels)):
                idx = labels == l
                model = self.model_map[l]
                if method == 'predict_proba':
                    # predict classification probabilities
                    temp_proba = model.predict_proba(X[idx])
                    for c, p in zip(model.classes_, temp_proba.T):
                        predictions[idx, classes_map[c]] = p
                else:
                    predictions[idx] = model.predict(X[idx])

        return predictions

    def fit(self, X, y, **kwargs):
        super().fit(X, y, **kwargs)
        self.post_fit(X, y)

    def predict(self, X, check_input=True):
        labels = super().apply(X)
        return self.post_predict(X, labels)


class RandomWeightRidge(Ridge):

    def __init__(self, alpha=1.0, *, fit_intercept=True, normalize="deprecated", copy_X=True, max_iter=None, tol=1e-3,
                 solver="auto", positive=False, random_state=None, initial_weight=None):
        super().__init__(alpha, fit_intercept=fit_intercept, normalize=normalize, copy_X=copy_X, max_iter=max_iter,
                         tol=tol, solver=solver, positive=positive, random_state=random_state)
        self.initial_weight = initial_weight

    def fit(self, X, y, sample_weight=None):
        return super().fit(X, y, np.abs(X[:, -1]) if np.sum(np.abs(X[:, -1])) != 0 else None)


class PLTreeRegressor(DecisionTreeRegressor, PLTree):
    """
    A simple implementation of piecewise linear regression tree
    """

    def __init__(self, base_model='Ridge', **kwargs):
        """
        base_model: The local model
        """
        self.base_model = base_model
        super().__init__(**kwargs)

    def model_controller(self):
        if self.base_model == 'RidgeCV':
            cv = RidgeCV()
        elif self.base_model == 'Ridge':
            cv = Ridge()
        elif self.base_model == 'LR':
            cv = LinearRegression()
        else:
            raise Exception
        return cv


class RPLBaseModel(BaseEstimator):
    base_model: Callable
    ridge: Union[LinearModel, LogisticRegression]

    def __init__(self, decision_tree_count, max_leaf_nodes):
        self.decision_tree_count = decision_tree_count
        self.max_leaf_nodes = max_leaf_nodes
        self.regression = isinstance(self, RegressorMixin)
        self.dt: List[Union[DecisionTreeRegressor, DecisionTreeClassifier]] = \
            [self.base_model(max_leaf_nodes=max_leaf_nodes, min_samples_leaf=10) if self.regression else
             # For classification cases
             MultiOutputRegressor(self.base_model(max_leaf_nodes=max_leaf_nodes, min_samples_leaf=10))
             for _ in range(self.decision_tree_count)]

    def fit(self, X, y):
        self.ridge.fit(X, y)
        if self.regression:
            prediction = self.ridge.predict(X)
        else:
            y = OneHotEncoder(sparse=False).fit_transform(y.reshape(-1, 1))
            prediction = self.ridge.predict_proba(X)
        residual = (y - prediction)
        for dt in self.dt:
            dt.fit(X, residual)
            if self.regression:
                prediction += dt.predict(X)
                residual = (y - prediction)
            else:
                prediction += dt.predict(X)
                p = softmax(prediction)
                residual = (y - p)

        # Calculate feature importance values
        # Feature importance in the global model
        self.feature_importance = np.abs(self.ridge.coef_)
        if len(self.feature_importance.shape) == 2:
            self.feature_importance = np.max(self.feature_importance, axis=0)
        self.feature_importance = self.feature_importance / self.feature_importance.sum()
        for dt in self.dt:
            # It is possible that decision tree not uses any features.
            # In this case, feature importance values will be zero.
            # assert np.isclose(np.sum(dt.feature_importances_), 1)
            if hasattr(dt, 'feature_importance'):
                self.feature_importance = np.max([dt.feature_importance, self.feature_importance], axis=0)
            elif hasattr(dt, 'feature_importances_'):
                self.feature_importance = np.max([dt.feature_importances_, self.feature_importance], axis=0)
            elif isinstance(dt, MultiOutputRegressor):
                feature_importance = np.max([d.feature_importances_ for d in dt.estimators_], axis=0)
                self.feature_importance = np.max([feature_importance, self.feature_importance], axis=0)
            else:
                raise Exception
        self.feature_importance /= self.feature_importance.sum()
        self.feature_importance = np.nan_to_num(self.feature_importance)
        assert np.all(self.feature_importance >= 0)
        return self

    def predict_proba(self, X):
        prediction = self.ridge.predict_proba(X)
        if np.isnan(prediction).any():
            raise Exception
        # assert not np.isnan(prediction).any(), f"Coefficient of LR: {self.ridge.coef_},{self.ridge.intercept_},{prediction},{X}"
        for dt in self.dt:
            prediction += dt.predict(X)
        prediction = softmax(prediction, axis=1)
        return prediction

    def predict(self, X):
        if not self.regression:
            return np.argmax(self.predict_proba(X))
        prediction = self.ridge.predict(X)
        for dt in self.dt:
            prediction += dt.predict(X)
        return prediction


class RidgeDT(RPLBaseModel, RegressorMixin):
    def __init__(self, decision_tree_count=1, max_leaf_nodes=4):
        if max_leaf_nodes is None:
            max_leaf_nodes = 4
        self.ridge = Ridge()
        self.base_model = PLTreeRegressor
        super().__init__(decision_tree_count, max_leaf_nodes)


class LRDTClassifier(RPLBaseModel, ClassifierMixin):
    def __init__(self, decision_tree_count=1, max_leaf_nodes=4):
        if max_leaf_nodes is None:
            max_leaf_nodes = 4
        self.ridge = LogisticRegression(max_iter=1000, solver='liblinear')
        self.base_model = PLTreeRegressor
        super().__init__(decision_tree_count, max_leaf_nodes)

    def fit(self, X, y):
        self.classes_ = np.unique(y)
        return super().fit(X, y)


class RidgeDTPlus(RidgeDT):
    """
    Using a decision tree regressor to further improve the performance.
    """

    def __init__(self, decision_tree_count=0, max_leaf_nodes=4,
                 final_model_splitter='random'):
        super().__init__(decision_tree_count, max_leaf_nodes)
        self.final_model_splitter = final_model_splitter
        self.dt.append(DecisionTreeRegressor(splitter=final_model_splitter))


class PLTreeClassifier(DecisionTreeClassifier, PLTree):
    """
    A simple implementation of piecewise linear regression tree
    """

    def __init__(self, *, criterion="gini", splitter="best", max_depth=None, min_samples_split=2, min_samples_leaf=1,
                 min_weight_fraction_leaf=0., max_features=None, random_state=None, max_leaf_nodes=None,
                 min_impurity_decrease=0., min_impurity_split=None, class_weight=None, ccp_alpha=0.0,
                 base_model='LR'):
        self.base_model = base_model
        super().__init__(criterion=criterion, splitter=splitter, max_depth=max_depth,
                         min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
                         min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features,
                         random_state=random_state, max_leaf_nodes=max_leaf_nodes,
                         min_impurity_decrease=min_impurity_decrease, min_impurity_split=min_impurity_split,
                         class_weight=class_weight, ccp_alpha=ccp_alpha)

    def model_controller(self):
        if self.base_model == 'LR':
            cv = LogisticRegression(solver='liblinear')
        elif isinstance(self.base_model, ClassifierMixin):
            cv = self.base_model
        else:
            raise Exception
        return cv

    def predict_proba(self, X, check_input=True):
        labels = super().apply(X)
        return self.post_predict(X, labels, method='predict_proba')


class SoftPLTreeRegressor(RegressorMixin, PLTree):
    """
    A simple implementation of piecewise linear regression tree
    """

    def __init__(self, *, criterion="gini", splitter="best", max_depth=None, min_samples_split=2, min_samples_leaf=1,
                 min_weight_fraction_leaf=0., max_features=None, random_state=None, max_leaf_nodes=None,
                 min_impurity_decrease=0., min_impurity_split=None, class_weight=None, ccp_alpha=0.0,
                 base_model='RidgeCV', feature_num=None, gene_num=None,
                 only_constructed_features=True, only_original_features=True,
                 partition_model='DecisionTree', partition_number=1):
        """
        :param feature_num: Number of all features
        :param gene_num: Number of constructed features
        """
        self.base_model = base_model
        self.feature_num = feature_num
        self.gene_num = gene_num
        self.importance_value = None
        self.only_constructed_features = only_constructed_features
        self.only_original_features = only_original_features
        self.partition_model = partition_model
        self.partition_number = partition_number
        parameters = dict(criterion=criterion, splitter=splitter, max_depth=max_depth,
                          min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
                          min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features,
                          random_state=random_state, max_leaf_nodes=max_leaf_nodes,
                          min_impurity_decrease=min_impurity_decrease, min_impurity_split=min_impurity_split,
                          class_weight=class_weight, ccp_alpha=ccp_alpha)
        if partition_model == 'DecisionTree':
            self.partition_tree = DecisionTreeClassifier(**parameters)
        elif partition_model == 'DecisionTree-Regression':
            self.partition_tree = DecisionTreeRegressor(max_leaf_nodes=partition_number)
        elif partition_model == 'LogisticRegression':
            self.partition_tree = LogisticRegression(solver='liblinear')
        elif partition_model == 'GMM':
            self.partition_tree = GaussianMixture(n_components=partition_number)
        elif partition_model == 'K-Means':
            self.partition_tree = KMeans(n_clusters=partition_number)
        else:
            raise Exception
        super().__init__(**parameters)

    def model_controller(self):
        if self.base_model == 'RidgeCV':
            cv = RidgeCV()
        elif self.base_model == 'RidgeCV-Log':
            cv = RidgeCV(alphas=(0.1, 1, 10, 100, 1000))
        elif self.base_model == 'Ridge':
            cv = Ridge()
        elif self.base_model == 'LR':
            cv = LinearRegression()
        elif isinstance(self.base_model, ClassifierMixin):
            cv = self.base_model
        else:
            raise Exception
        return cv

    def complexity(self):
        # count decision tree nodes
        count = 0
        count += self.partition_tree.tree_.node_count
        # count local model coefficients
        for k, v in self.model_map.items():
            v: LinearRegression
            count += len(v.coef_)
        return count

    def fit(self, X, y, sample_weight=None, check_input=True,
            X_idx_sorted="deprecated"):
        X, label = X[:, :-1], X[:, -1]
        label = LabelEncoder().fit_transform(label)
        assert X.shape[1] == self.feature_num
        if isinstance(self.partition_tree, DecisionTreeClassifier):
            if X.shape[1] > self.gene_num and self.only_original_features:
                # construct the space partition tree with original features
                self.partition_tree.fit(X[:, self.gene_num:], label, sample_weight=sample_weight,
                                        check_input=check_input,
                                        X_idx_sorted=X_idx_sorted)
                self.importance_value = np.zeros(self.gene_num)
            else:
                # construct the space partition tree with constructed features
                self.partition_tree.fit(X, label, sample_weight=sample_weight, check_input=check_input,
                                        X_idx_sorted=X_idx_sorted)
                self.importance_value = self.partition_tree.feature_importances_
        elif isinstance(self.partition_tree, DecisionTreeRegressor):
            self.partition_tree.fit(X, label)
            self.importance_value = self.partition_tree.feature_importances_
        elif isinstance(self.partition_tree, GaussianMixture):
            self.partition_tree.fit(X)
            self.importance_value = np.zeros(X.shape[1])
            self.partition_tree.classes_ = self.partition_tree.predict(X)
        elif isinstance(self.partition_tree, KMeans):
            label = self.partition_tree.fit_predict(X)
            self.partition_tree = DecisionTreeClassifier()
            self.partition_tree.fit(X, label, sample_weight=sample_weight, check_input=check_input,
                                    X_idx_sorted=X_idx_sorted)
            self.importance_value = self.partition_tree.feature_importances_
            self.partition_tree.classes_ = self.partition_tree.predict(X)
        else:
            self.partition_tree.fit(X, label)
            self.importance_value = np.max(self.partition_tree.coef_, axis=0)
        if isinstance(self.partition_tree, (ClassifierMixin, BaseMixture)):
            if X.shape[1] > self.gene_num and self.only_original_features:
                partition_scheme = self.partition_tree.predict_proba(X[:, self.gene_num:])
            else:
                partition_scheme = self.partition_tree.predict_proba(X)
        elif isinstance(self.partition_tree, DecisionTreeRegressor):
            partition_scheme = self.partition_tree.apply(X)
        else:
            raise Exception
        # training final results
        if self.only_constructed_features:
            X = X[:, :self.gene_num]
        self.post_fit(X, y, partition_scheme=partition_scheme)

    def predict(self, X, check_input=True):
        labels_prob, predict_values = self.basic_prediction(X)
        predict_values = np.sum(predict_values * labels_prob, axis=1)
        return predict_values

    def basic_prediction(self, X):
        X = X[:, :self.feature_num]
        if X.shape[1] > self.gene_num and self.only_original_features:
            # get the space partition results with original features
            labels = self.partition_tree.predict(X[:, self.gene_num:])
            # probability matrix (size: samples * clusters)
            labels_prob = self.partition_tree.predict_proba(X[:, self.gene_num:])
        else:
            # get the space partition results with constructed features
            labels = self.partition_tree.predict(X)
            # probability matrix (size: samples * clusters)
            labels_prob = self.partition_tree.predict_proba(X)
        if self.only_constructed_features:
            X = X[:, :self.gene_num]
        # prediction result matrix (size: samples * clusters)
        predict_values = self.post_predict(X, labels, method='soft_prediction',
                                           classes=self.partition_tree.classes_)
        return labels_prob, predict_values

    def score(self, X, y, sample_weight=None):
        # best partition scheme
        _, predict_values = self.basic_prediction(X)
        # regret = (((np.sum((predict_values * self.partition_tree.predict_proba(X[:, :-1])), axis=1)
        #             - y.flatten()) ** 2 - np.min((predict_values - y) ** 2, axis=1)))
        # print('Max Regret', np.max(regret))
        return np.argmin((predict_values - y) ** 2, axis=1).astype(int)

    def em_algorithm(self, X, y, random_initialization=False):
        if random_initialization:
            partition_scheme = np.random.randint(0, self.partition_number, len(X))
        else:
            partition_scheme = Pipeline([
                ('StandardScaler', StandardScaler()),
                # ('K-Means', KMeans(self.partition_number)),
                # ('K-Means', SpectralClustering(self.partition_number)),
                ('K-Means', DBSCAN()),
            ]).fit_predict(X)
            print(np.unique(partition_scheme))

        inconsistent_partition = np.inf
        iteration = 0
        max_iteration = 20
        cv_em = False
        while inconsistent_partition >= 10:
            if iteration >= max_iteration:
                break
            cv = KFold(n_splits=5, shuffle=True, random_state=0)
            if cv_em:
                # determine the best partition scheme through cross-validation
                new_partition_scheme = np.zeros(len(y))
                for index in cv.split(X):
                    train_index, test_index = index
                    SoftPLTreeRegressor.fit(self, np.concatenate([X[train_index],
                                                                  np.reshape(partition_scheme[train_index], (-1, 1))],
                                                                 axis=1),
                                            y[train_index])
                    new_partition_scheme[test_index] = SoftPLTreeRegressor.score(self, X[test_index],
                                                                                 y[test_index].reshape(-1, 1))
            else:
                SoftPLTreeRegressor.fit(self, np.concatenate([X, np.reshape(partition_scheme, (-1, 1))], axis=1), y)
                new_partition_scheme = SoftPLTreeRegressor.score(self, X, y.reshape(-1, 1))
            inconsistent_partition = np.sum(new_partition_scheme != partition_scheme)
            print('iteration', iteration, 'inconsistent partition', inconsistent_partition)
            partition_scheme = new_partition_scheme
            iteration += 1
        return iteration

    @property
    def feature_importances_(self):
        return self.importance_value


class SoftPLTreeRegressorEM(SoftPLTreeRegressor):
    def fit(self, X, y, sample_weight=None, check_input=True, X_idx_sorted="deprecated"):
        super().em_algorithm(X, y)


def regression_task_demo():
    X, y = load_diabetes(return_X_y=True)
    X, y = np.array(X), np.array(y)
    x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)
    for i in range(10):
        lr = BaggingRegressor(RidgeDTPlus(decision_tree_count=i), n_estimators=100)
        lr.fit(x_train, y_train)
        print('Training Score', r2_score(lr.predict(x_train), y_train))
        print('Testing Score', r2_score(lr.predict(x_test), y_test))


def classification_task_demo():
    X, y = load_wine(return_X_y=True)
    X, y = np.array(X), np.array(y)
    x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)
    for dt_count in [0, 1, 2]:
        lr = LRDTClassifier(decision_tree_count=dt_count)
        lr.fit(x_train, y_train)
        print('Training Score', roc_auc_score(y_train, lr.predict_proba(x_train), average='macro',
                                              multi_class='ovo'))
        print('Testing Score', roc_auc_score(y_test, lr.predict_proba(x_test), average='macro',
                                             multi_class='ovo'))


if __name__ == '__main__':
    regression_task_demo()
    # classification_task_demo()



import random

import numpy as np
from deap.gp import mutUniform
from scipy.stats import pearsonr, spearmanr

from evolutionary_forest.component.evaluation import quick_evaluate


def varAndPlus(population, toolbox, cxpb, mutpb, gene_num, limitation_check,
               semantic_check_tool=None, varOr=False):
    @limitation_check
    def mutation_function(*population):
        offspring = [toolbox.clone(ind) for ind in population]
        crossed_individual = set()

        # Apply crossover and mutation on the offspring
        # Support both VarAnd and VarOr
        i = 0
        while i < len(offspring):
            # Execute mutation and selection operator N-times
            for c in range(gene_num):
                if i % 2 == 0 and random.random() < cxpb:
                    offspring[i], offspring[i + 1] = toolbox.mate(offspring[i], offspring[i + 1])
                    del offspring[i].fitness.values, offspring[i + 1].fitness.values
                    crossed_individual.add(offspring[i])
                    crossed_individual.add(offspring[i+1])

                if random.random() < mutpb and (not varOr or (i not in crossed_individual)):
                    offspring[i], = toolbox.mutate(offspring[i])
                    del offspring[i].fitness.values

            if semantic_check_tool is not None:
                # check by semantics
                x = semantic_check_tool['x']
                y = semantic_check_tool['y']
                pset = semantic_check_tool['pset']
                correlation_threshold=semantic_check_tool.get('correlation_threshold', 0.2)
                correlation_mode=semantic_check_tool.get('correlation_mode', 'Pearson')
                index = np.random.randint(0, len(x), 20)
                for k, g in enumerate(offspring[i].gene):
                    y_hat = quick_evaluate(g, pset, x[index])
                    c = 0
                    function = {
                        'Pearson': pearsonr,
                        'Spearman': spearmanr,
                    }[correlation_mode]
                    while (not isinstance(y_hat, np.ndarray)) or (y_hat.size != y[index].size) or \
                        (np.abs(function(y_hat, y[index])[0]) < correlation_threshold):
                        c += 1
                        offspring[i].gene[k] = mutUniform(g, toolbox.expr_mut, pset)[0]
                        del offspring[i].fitness.values
                        y_hat = quick_evaluate(g, pset, x[index])
                        if c > 100:
                            print('Warning!')
                            break
            i += 1
        return offspring

    return mutation_function(*population)



import numpy as np
from sklearn.preprocessing import StandardScaler


def _handle_zeros_in_scale(scale, copy=True):
    """Makes sure that whenever scale is zero, we handle it correctly.

    This happens in most scalers when we have constant features.
    """
    threshold = 1e-10

    # if we are fitting on 1D arrays, scale might be a scalar
    if np.isscalar(scale):
        if scale <= threshold:
            scale = 1.
        return scale
    elif isinstance(scale, np.ndarray):
        if copy:
            # New array to avoid side-effects
            scale = scale.copy()
        scale[scale <= threshold] = 1.0
        return scale


class SafetyScaler(StandardScaler):
    """
    This is a scaler to eliminate negligible variance
    """

    def partial_fit(self, X, y=None, sample_weight=None):
        super().partial_fit(X, y, sample_weight)
        if self.with_std:
            self.scale_ = _handle_zeros_in_scale(np.sqrt(self.var_))
        else:
            self.scale_ = None

        return self

    def transform(self, X, copy=None):
        return np.nan_to_num(super().transform(X, copy), posinf=0, neginf=0)


if __name__ == '__main__':
    data = np.array([[0, 0], [0, 0], [1e-50, 1], [1, 1]])
    # data = np.random.randn(100, 100) * 100
    scaler = SafetyScaler()
    scaler.fit(data[:3, ])
    print(scaler.transform(data[3:, ]))
